import inspect
import typing
from binascii import a2b_hex
from typing import Union

import pytest

from smllib import SmlStreamReader
from smllib import crc as crc_module
from smllib.crc.x25 import get_crc as x25_get_crc


@pytest.mark.parametrize('msg', (
    pytest.param('7605032b18116200620072630201710163fa3600', id='crc1'),
    pytest.param('7605171756bd6200620072630201710163170f00', id='crc2'),
    pytest.param('760700070615fce7620062007263010176010107000702e053ce0b0901454d4800004895f701016333aa00', id='crc3'),
    pytest.param('76051c414c02620062007263010176010102310b0a01445a47000282c0b07262016505471c2a620263f93800', id='crc4'),
    pytest.param('76040000016200620072650000010176010107000002dba23c0b0a01484c5902000424a0010163945b00', id='crc5'),
))
def test_crc_x25(msg) -> None:
    _msg = memoryview(a2b_hex(msg))
    crc_msg = f'{_msg[-3]:02x}{_msg[-2]:02x}'
    crc_calc = f'{x25_get_crc(_msg[:-4]):04x}'
    assert crc_msg == crc_calc


def _get_signature() -> inspect.Signature:
    return inspect.Signature(
        parameters=[
            inspect.Parameter('buf', kind=inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Union[memoryview, bytes])
        ],
        return_annotation=int
    )


@pytest.mark.parametrize('name', (n for n in dir(crc_module) if not n.startswith('_')))
def test_signature_crc_funcs(name: str) -> None:
    crc_impl = getattr(crc_module, name)
    crc_sig = inspect.signature(crc_impl.get_crc)
    assert crc_sig == _get_signature()


def test_type_hint_reader() -> None:
    # Literal
    available = [n for n in dir(crc_module) if not n.startswith('_')]
    hint = typing.get_type_hints(SmlStreamReader.__init__)
    literal = hint['crc']
    literal_values = typing.get_args(literal)
    assert set(available) == set(literal_values)

    # crc_func variable
    signature = _get_signature()
    a = SmlStreamReader()
    hint = typing.get_type_hints(a.crc_func)
    assert hint.pop('return') is signature.return_annotation

    assert hint
    for name in hint:
        assert hint[name] == signature.parameters[name].annotation


def test_invalid_crc_name() -> None:

    with pytest.raises(ValueError) as e:
        SmlStreamReader(crc='asfd')
    assert str(e.value) == 'Unsupported CRC "asfd"! Available: "kermit", "x25"'



import typing
from typing import Final, TypeVar


INDENT: Final = '  '


class SmlBaseObj:
    __sml__: typing.ClassVar[dict]

    def format_msg(self, indent: int = 0) -> str:
        indent += 1
        r = f'<{self.__class__.__name__}>\n'
        w = max(map(len, self.__dict__), default=0)
        for k, v in self.__dict__.items():
            if isinstance(v, SmlBaseObj):
                r += f'{INDENT * indent}{k!s:s} {v.format_msg(indent):s}'
            elif isinstance(v, (tuple, list)):
                r += f'{INDENT * indent}{k!s:s}:\n'
                for e in v:
                    r += f'{INDENT * (indent + 1)}{e.format_msg(indent + 1):s}'
            else:
                r += f'{INDENT * indent}{k!s:{w}s}: {v}\n'
        return r


T_SML_OBJ = TypeVar('T_SML_OBJ', bound=SmlBaseObj, covariant=True)



from smllib.sml._base_obj import T_SML_OBJ, SmlBaseObj
from smllib.sml.sml_eom import EndOfSmlMsg


# isort: split


from smllib.sml.sml_choice import SmlChoice


# isort: split


from smllib.sml._field_info import SmlObjFieldInfo, inspect_obj


# isort: split


from smllib.sml.list_entry import SmlListEntry
from smllib.sml.response_get_list import SmlGetListResponse
from smllib.sml.response_open_close import SmlCloseResponse, SmlOpenResponse
from smllib.sml.sml_obis import ObisCode


# isort: split


from smllib.sml.message import SmlMessage



from ._builder import T_SML_BUILDER, SmlObjBuilder


# isort: split


from .list_entry import SmlListEntryBuilder
from .message import SmlMessageBuilder
from .response_get_list import SmlGetListResponseBuilder
from .response_open_close import SmlCloseResponseBuilder, SmlOpenResponseBuilder


# isort: split


from .default_context import CTX_HINT, create_context



from smllib.errors import UnsupportedChoiceValue
from smllib.sml.sml_time import build_time


def build_sml_value(_in):
    if _in is None:
        return _in

    if not isinstance(_in, list):
        return _in

    type_s, value_s = _in
    _type = type_s.value
    if _type == 1:  # SML_Time
        return build_time(value_s.value)

    raise UnsupportedChoiceValue(_type)



from binascii import a2b_hex

from smllib.sml_frame import EndOfSmlMsg, SmlFrame, SmlFrameSnippet


def check(f: SmlFrameSnippet, value, msg: str) -> None:
    if value is True or value is False or value is None or value is EndOfSmlMsg:
        assert f.value is value
    else:
        if isinstance(value, (tuple, list)):
            for i, _value in enumerate(value):
                assert f.value[i].value == _value
        else:
            assert f.value == value
    assert f.msg.hex() == msg


def test_get_int8() -> None:
    f = SmlFrame(b'\x52\xff')
    check(f.get_value(0), -1, '52ff')
    assert f.next_pos == 2

    f = SmlFrame(b'\x52\x03')
    check(f.get_value(0), 3, '5203')
    assert f.next_pos == 2

    f = SmlFrame(b'\x00\x52\x03')
    f.next_pos = 9999
    check(f.get_value(1), 3, '5203')
    assert f.next_pos == 3


def test_get_uint8() -> None:
    f = SmlFrame(b'\x62\xff')
    check(f.get_value(0), 255, '62ff')
    assert f.next_pos == 2

    f = SmlFrame(b'\x62\x03')
    check(f.get_value(0), 3, '6203')
    assert f.next_pos == 2

    f = SmlFrame(b'\x00\x62\xff')
    f.next_pos = 9999
    check(f.get_value(1), 255, '62ff')
    assert f.next_pos == 3


def test_get_int16() -> None:
    f = SmlFrame(b'\x53\xff\x00')
    check(f.get_value(0), -256, '53ff00')
    assert f.next_pos == 3

    f = SmlFrame(b'\x53\x00\x03')
    check(f.get_value(0), 3, '530003')
    assert f.next_pos == 3

    f = SmlFrame(b'\x00\x53\xff\x01')
    f.next_pos = 9999
    check(f.get_value(1), -255, '53ff01')
    assert f.next_pos == 4


def test_get_uint16() -> None:
    f = SmlFrame(b'\x63\xff\x00')
    check(f.get_value(0), 65280, '63ff00')
    assert f.next_pos == 3

    f = SmlFrame(b'\x63\x00\x03')
    check(f.get_value(0), 3, '630003')
    assert f.next_pos == 3

    f = SmlFrame(b'\x00\x63\xff\x01')
    f.next_pos = 9999
    check(f.get_value(1), 65281, '63ff01')
    assert f.next_pos == 4


def test_get_int32() -> None:
    f = SmlFrame(b'\x55\x00\x00\x0a\x8c')
    check(f.get_value(0), 2700, '5500000a8c')
    assert f.next_pos == 5

    f = SmlFrame(b'\x55\xFF\xFF\xFF\xFF')
    check(f.get_value(0), -1, '55ffffffff')
    assert f.next_pos == 5

    f = SmlFrame(b'\x01\x01\x55\x00\x00\x0a\x8c')
    f.next_pos = 9999
    check(f.get_value(2), 2700, '5500000a8c')
    assert f.next_pos == 7


def test_get_uint32() -> None:
    f = SmlFrame(b'\x65\x00\x00\x0a\x8c')
    check(f.get_value(0), 2700, '6500000a8c')
    assert f.next_pos == 5

    f = SmlFrame(b'\xab\xcd\x87\x44\x65\x0c\x6a\x50\xb5')
    check(f.get_value(4), 208294069, '650c6a50b5')
    assert f.next_pos == 9


def test_none() -> None:
    f = SmlFrame(b'\x01\x01\x01')
    assert f.get_value().value is None
    assert f.next_pos == 1

    f.next_pos = 9999
    assert f.get_value(1).value is None
    assert f.next_pos == 2

    assert f.get_value().value is None
    assert f.next_pos == 3


def test_get_list() -> None:
    f = SmlFrame(b'\x71\x01')
    f.next_pos = 9999
    check(f._parse_msg(f.get_value(0)), [None], '7101')
    assert f.next_pos == 2

    f = SmlFrame(b'\x72\x01\x00')
    f.next_pos = 9999
    check(f._parse_msg(f.get_value(0)), [None, EndOfSmlMsg], '720100')
    assert f.next_pos == 3

    f = SmlFrame(b'\x72\x52\xff\x62\x1e')
    check(f._parse_msg(f.get_value(0)), [-1, 30], '7252ff621e')
    assert f.next_pos == 5

    f = SmlFrame(a2b_hex(
        '77070100010800ff'
        '6500000782'
        '01'
        '621e'
        '52ff'
        '59000000000dd359d6'
        '01'
        'ff'))
    check(f._parse_msg(f.get_value(0)),
          ['0100010800ff', 1922, None, 30, -1, 231954902, None],
          '77070100010800ff650000078201621e52ff59000000000dd359d601')
    assert f.next_pos == 28


def test_str() -> None:
    f = SmlFrame(b'\x07\x01\x00\x01\x08\x00\xFF')
    f.next_pos = 9999
    check(f.get_value(0), '0100010800ff', '070100010800ff')
    assert f.next_pos == 7


def test_long_str() -> None:
    f = SmlFrame(a2b_hex('8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102'))  # noqa: E501
    check(
        f.get_value(0),
        '010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102',
        '8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102')
    assert f.next_pos == 50

    f = SmlFrame(a2b_hex('8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102FF'))  # noqa: E501
    check(
        f.get_value(0),
        '010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102',
        '8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102')
    assert f.next_pos == 50
    assert f.buffer[f.next_pos] == 0xFF


def test_bool() -> None:
    f = SmlFrame(b'\x42\x01')
    f.next_pos = 9999
    check(f.get_value(0), True, '4201')
    assert f.next_pos == 2

    f = SmlFrame(b'\x42\x00')
    check(f.get_value(0), False, '4200')
    assert f.next_pos == 2


def test_eom() -> None:
    f = SmlFrame(b'\x00')
    check(f.get_value(0), EndOfSmlMsg, '00')
    assert f.next_pos == 1

    f = SmlFrame(b'\x00\x00')
    check(f.get_value(1), EndOfSmlMsg, '00')
    assert f.next_pos == 2
